"""
Visualizes the landscape of a protein.
"""

import matplotlib.pyplot as plt
import matplotlib.ticker as tck
from matplotlib.ticker import MaxNLocator
import pandas as pd 
import numpy as np
import logging
logger = logging.getLogger(__name__)


def plot_peptide_length_dist(first_df: pd.DataFrame, second_df: pd.DataFrame, first_explode: str, second_explode: str, first_column: str, second_column: str, first_label: str, second_label: str) -> tuple[plt.figure,int,int]:
    """Visualize the distribution of lengths of sequences.
    
    Args:
        first_df: A pandas dataframe.
        second_df: A pandas dataframe.
        first_explode: The header of the column, that is used to explode the 
            dataframe. 
        second_column: The header of the column, that is used to explode the 
            dataframe. 
        first_column: The header of the column, that holds the sequences in 
            first_df, for which the length distribution will be plotted. 
        second_column: The header of the column, which holds the sequences in 
            second_df, for which the length distribution will be plotted.  
        first_label: Label for the values of first_column in the plot.
        second_label: Label for the values of second_column in the plot. 

    Returns:
        A tuple including a matplotlib figure and two integers. The figure is a 
        histogram visualizing the length distribution of peptides and epitopes. 
        The two integers are the number of peptides and the number of epitopes. 
    """
    
    first_long = first_df.explode(first_explode)
    second_long = second_df.explode(second_explode)
    
    logger.info(f'{len(first_long)} peptides were reduced to {len(second_long)} epitopes.')

    # compute a histogram of the sequence lengths in first_column and second_column
    fig, ax = plt.subplots(layout='constrained')
    seq_first = first_long[first_column]
    seq_second = second_long[second_column]

    first_len = seq_first.map(lambda pep: len(pep)).to_list()
    second_len = seq_second.map(lambda pep: len(pep)).to_list()
    
    ax.hist(first_len, bins=np.arange(5,50,1), color='grey', label=first_label, alpha=0.6)
    ax.hist(second_len, bins=np.arange(5,50,1), color='red', label=second_label, alpha=0.6)

    ax.legend()
    ax.set_xlabel('length')
    ax.set_ylabel('count')
    return fig, len(first_long), len(second_long)


def plot_protein_landscape(protein_df: pd.DataFrame, accession: str, proteome_dict: dict[str,str]) -> plt.figure:
    """Visualize the landscape of a protein.

    Args:
        protein_df: A pandas dataframe containing one protein per row.
        accession: The string of a protein accession.
        proteome_dict: A dictionary containing the reference proteome.

    Returns:
        A matplotlib bar plot that visualizes the peptide and core epitope 
        distribution across the sequence of the protein with the provided 
        accession. 
    """

    prot_row = protein_df[(protein_df['accession'] == accession)]
    if len(prot_row) == 0:
        raise Exception('The accession {} is not in your input data.'.format(accession))

    prot_seq = proteome_dict[accession]

    prot_landscape = [0 for _ in prot_seq]
    
    fig_width = max(1,round(len(prot_landscape)/50))
    fig_height = 3

    max_intens = 0

    fig, ax = plt.subplots(figsize=(fig_width, fig_height), layout='constrained')

    for group, landscape in enumerate(prot_row['landscape'].iloc[0]):

        if group % 3 == 0:
            color = 'red'
        elif group % 3 == 1:
            color = 'blue'
        else: 
            color = 'green'

        group_start = min(prot_row['grouped_peptides_start'].iloc[0][group])
        for idx, position in enumerate(landscape):
            ax.bar(group_start+int(idx),position,width=1, alpha=0.4, color=color)
        for pos in range(prot_row['core_epitopes_start'].iloc[0][group],prot_row['core_epitopes_end'].iloc[0][group]):
            ax.bar(pos,0.5,width=1,color=color)

        max_intens = max(max_intens, max(landscape))

    ybins=max_intens/10
    ax.yaxis.set_major_locator(MaxNLocator(nbins=max(ybins, 5)))
    ax.xaxis.set_major_locator(MaxNLocator(nbins=max(fig_width/10, 15)))

    ax.set_title('Number of peptides mapped to each amino acid position and core epitopes of protein {}'.format(accession))
    ax.set_xlabel('Position in protein {}'.format(accession))
    ax.set_ylabel('Number of mapped peptides')
    return fig



def plot_core_mapping_peptides_hist(epitope_df: pd.DataFrame) ->  plt.figure:
    """A histogram of the number of peptides mapped to each epitope. 
    
    Args:
        epitope_df: A dataframe containing one epitope and 
            information of that epitope per row. 

    Returns:
        A histogram visualizing the number of peptides mapped to 
        each epitope. 
    """
    fig, ax = plt.subplots(layout='constrained')
    n_peps = epitope_df['grouped_peptides_sequence'].apply(lambda sequences: len(sequences.split(',')))
    ax.hist(n_peps,bins=np.arange(1,max(n_peps)+1,1))
    ax.set_yscale('log')
    ax.set_xlabel('number of peptides contributing to epitope')
    ax.set_ylabel('count')
    return fig





"""
Computes core epitopes, by grouping overlapping peptides, together, building a landscape for each group and identifying plateaus with a defined minimal length in each landscape. 
"""

import numpy as np
import re
import pandas as pd


def group_peptides(protein_df: pd.DataFrame, min_overlap: int, max_step_size: int, intensity_column: str) -> pd.DataFrame:
    """Group the peptides to consensus epitopes.

    Args:
        protein_df: A pandas dataframe containing one protein per row.
        min_overlap: An integer of the minimal overlap between two epitopes     
            to be grouped to the same consensus epitope.
        max_step_size: An integer of the maximal distance between the start 
            position of two epitopes to be grouped to the same consensus 
            epitope.
        intensity_column: The header of the column containing the intensities
            of the peptides.

    Returns:
        The protein_df with the five to seven additional columns: 
        grouped_peptides_start, grouped_peptides_end, 
        grouped_peptides_sequence, grouped peptides_intensity, 
        core_epitopes_intensity, relative_core_intensity and 
        sequence_group_mapping. The first five column contain the start
        position, end position, sequence and intensity of the peptides grouped
        to consensus epitopes. The column relative_core_intensity contains the 
        relative core intensities of the consensus core epitopes. The column 
        sequence group mapping contains for each peptide to which consensus 
        epitope it is grouped.
    """
    # start, end, sequence and intensity of peptides of one group grouped together
    protein_df['grouped_peptides_start'] = [[] for _ in range(len(protein_df))]
    protein_df['grouped_peptides_end'] = [[] for _ in range(len(protein_df))]
    protein_df['grouped_peptides_sequence'] = [[] for _ in range(len(protein_df))]
    if intensity_column:
        protein_df['grouped_peptides_intensity'] = [[] for _ in range(len(protein_df))]
        # for each peptide group the total and relative intensity of the entire group
        protein_df['core_epitopes_intensity'] = [[] for _ in range(len(protein_df))]
        protein_df['relative_core_intensity'] = [[] for _ in range(len(protein_df))]
    # for each peptide the index of its group
    protein_df['sequence_group_mapping'] = [[] for _ in range(len(protein_df))]

    total_file_intensity = 0
    for r,row in protein_df.iterrows():
        
        start_pos =  row['start']
        end_pos = row['end']
        sequences = row['sequence']
        if intensity_column:
            intensity = row['intensity']

        grouped_peptides_start = []
        grouped_peptides_end = []
        grouped_peptides_sequence = []
        if intensity_column:
            grouped_peptides_intensity = []
            core_intensity = 0
        n_jumps = 0
        mapping = []

        for i in range(len(start_pos)-1):

            grouped_peptides_start.append(int(start_pos[i]))
            grouped_peptides_end.append(int(end_pos[i]))
            grouped_peptides_sequence.append(sequences[i])
            if intensity_column:
                grouped_peptides_intensity.append(intensity[i])

            step_size = int(start_pos[i+1]) - int(start_pos[i])
            pep_length = int(end_pos[i]) - int(start_pos[i])
            mapping.append(n_jumps)
            if intensity_column:
                core_intensity += float(intensity[i])
                total_file_intensity += float(intensity[i])

            # create new peptide group after each jump
            if (step_size >= max_step_size) and (pep_length <= step_size + min_overlap):
                protein_df.at[r,'grouped_peptides_start'].append(grouped_peptides_start)
                protein_df.at[r,'grouped_peptides_end'].append(grouped_peptides_end)
                protein_df.at[r,'grouped_peptides_sequence'].append(grouped_peptides_sequence)
                if intensity_column:
                    protein_df.at[r,'grouped_peptides_intensity'].append(grouped_peptides_intensity)
                    protein_df.at[r,'core_epitopes_intensity'].append(core_intensity)

                n_jumps += 1
                grouped_peptides_end = []
                grouped_peptides_start = []
                grouped_peptides_sequence = []
                grouped_peptides_intensity = []
                if intensity_column:
                    core_intensity = 0

        # special case for last peptide match of protein
        if len(grouped_peptides_end) == 0:
            protein_df.at[r,'grouped_peptides_start'].append([int(start_pos[-1])])
            protein_df.at[r,'grouped_peptides_end'].append([int(end_pos[-1])])
            protein_df.at[r,'grouped_peptides_sequence'].append([sequences[-1]])
            if intensity_column:
                protein_df.at[r,'grouped_peptides_intensity'].append([intensity[-1]])
                protein_df.at[r,'core_epitopes_intensity'].append(intensity[-1])
                total_file_intensity += float(intensity[-1])
            mapping.append(n_jumps)
        else:
            grouped_peptides_start.append(int(start_pos[-1]))
            grouped_peptides_end.append(int(end_pos[-1]))
            grouped_peptides_sequence.append(sequences[-1])
            if intensity_column:
                grouped_peptides_intensity.append(intensity[-1])
                core_intensity += float(intensity[-1])
                total_file_intensity += float(intensity[-1])
                protein_df.at[r,'grouped_peptides_intensity'].append(grouped_peptides_intensity)
                protein_df.at[r,'core_epitopes_intensity'].append(core_intensity)
            protein_df.at[r,'grouped_peptides_start'].append(grouped_peptides_start)
            protein_df.at[r,'grouped_peptides_end'].append(grouped_peptides_end)
            protein_df.at[r,'grouped_peptides_sequence'].append(grouped_peptides_sequence)
            mapping.append(n_jumps)

        protein_df.at[r,'sequence_group_mapping'] = mapping
    if intensity_column:
        protein_df['relative_core_intensity'] = protein_df['core_epitopes_intensity'].apply(lambda x: [float(ints)/total_file_intensity for ints in x])
    if intensity_column:
        protein_df['core_epitopes_intensity_all'] = protein_df.apply(lambda row: [row['core_epitopes_intensity'][i] for i in row['sequence_group_mapping']],axis=1)
        protein_df['relative_core_intensity_all'] = protein_df.apply(lambda row: [row['relative_core_intensity'][i] for i in row['sequence_group_mapping']],axis=1)
    
    return protein_df


def gen_landscape(protein_df: pd.DataFrame, mod_pattern: str) -> pd.DataFrame:
    """Compute the landscape of consensus epitope groups.
    
    Args:
        protein_df: A pandas dataframe containing one protein per row.
        mod_pattern A comma separated string with delimiters for peptide
            modifications

    Returns:
        The protein_df with two additional columns (landscape, whole_epitopes).
        The column landscape contains the landscapes of the consensus epitopes 
        group. The height of the landscape of a consensus epitope group at a 
        position is the number of peptides that contain that position. The 
        column whole_epitopes contains the entire sequence of the consensus 
        epitope group.

    Raises:
        Exception: If the mappings are contradictory.
    """

    protein_df['landscape'] = [[] for _ in range(len(protein_df))]
    protein_df['whole_epitopes'] = [[] for _ in range(len(protein_df))]
    protein_df['whole_epitopes_all'] = [[] for _ in range(len(protein_df))]

    for r, row in protein_df.iterrows():
        #seen_pep={}
        for group, [pep_group_start, pep_group_end] in enumerate(zip(row['grouped_peptides_start'], row['grouped_peptides_end'])):
            seen_pep={}
            start_idx = pep_group_start[0]
            group_landscape = [0 for _ in range(max(pep_group_end)+1-min(pep_group_start))]#np.zeros(max(pep_group_end)+1-min(pep_group_start)) 
            
            for pep_idx, pep_pos in enumerate(zip(pep_group_start, pep_group_end)):
                pep_start = pep_pos[0]
                pep_end = pep_pos[1]

                current_seq = row['grouped_peptides_sequence'][group][pep_idx]
                pattern = r'\(.*?\)'
                current_seq = re.sub(pattern,"",current_seq)
                pattern = r'\[.*?\]'
                current_seq = re.sub(pattern,"",current_seq)
                if mod_pattern:
                    pattern = re.escape(mod_pattern.split(',')[0]) + r'.*?' + re.escape(mod_pattern.split(',')[1])
                    current_seq = re.sub(pattern,"",current_seq)

                if str(pep_start) + '-' + str(pep_end) in seen_pep:
                    seen_seq = seen_pep[str(pep_start) + '-' + str(pep_end)]
                    check_seen = [(seq == current_seq) for seq in seen_seq]
                    # current_seq not in list for repetitive regions (dont increase landscape value for repetitive regions)
                    seen_pep[str(pep_start) + '-' + str(pep_end)].append(current_seq)
                else: 
                    # if sequence and position not seen before add the sequence to the landscape
                    seen_pep[str(pep_start) + '-' + str(pep_end)] = [current_seq]
                    for pos in range(pep_start, pep_end+1):
                        group_landscape[pos-start_idx] += 1

            protein_df.at[r,'landscape'].append(group_landscape)

            # build whole group sequences
            consensus_seq = ''
            consensus_pos = []
            for sequence, sequence_pos in zip(row['grouped_peptides_sequence'][group], row['grouped_peptides_start'][group]):
                sequence = re.sub(r"\(.*?\)","",sequence)
                sequence = re.sub(r'\[.*?\]',"",sequence)
                if mod_pattern:
                    pattern = re.escape(mod_pattern.split(',')[0]) + r'.*?' + re.escape(mod_pattern.split(',')[1])
                    sequence = re.sub(pattern,"",sequence)
                sequence_pos = [i for i in range(sequence_pos, sequence_pos + len(sequence))]
                for aa, aa_pos in zip(sequence, sequence_pos):
                    if aa_pos not in consensus_pos:
                        consensus_seq += aa
                        consensus_pos.append(aa_pos)
                    else:
                        if consensus_seq[consensus_pos.index(aa_pos)] != aa:
                            raise Exception('Something with the mapping went wrong! If you provided the start and end column please try again without providing the column header for these columns.')
            for _ in row['grouped_peptides_sequence'][group]:
                protein_df.at[r,'whole_epitopes_all'].append(consensus_seq)
            protein_df.at[r,'whole_epitopes'].append(consensus_seq)
        
    return protein_df


def get_consensus_epitopes(protein_df: pd.DataFrame, min_epi_len: int) -> pd.DataFrame:
    """Compute the consensus epitope sequence of each consensus epitope group.
    
    Args:
        protein_df: A pandas dataframe containing one protein per row.
        min_epi_length: An integer of the minimal length of a consensus epitope.

    Returns:
        The protein_df with one additional column, that contains the consensus epitope sequence of each consensus epitope group.
    """
    protein_df['consensus_epitopes'] = [[] for _ in range(len(protein_df))]
    protein_df['consensus_epitopes_all'] = [[] for _ in range(len(protein_df))]
    protein_df['core_epitopes_start'] = [[] for _ in range(len(protein_df))]
    protein_df['core_epitopes_end'] = [[] for _ in range(len(protein_df))]
    protein_df['core_epitopes_start_all'] = [[] for _ in range(len(protein_df))]
    protein_df['core_epitopes_end_all'] = [[] for _ in range(len(protein_df))]

    for r, row in protein_df.iterrows():
        for group,landscape in enumerate(row['landscape']):
            
            # build consensus epitopes
            total_counts = np.unique(landscape)
            total_counts[::-1].sort()
        
            # find total coverage for which consensus epitope is at least min_epi_len long
            for total_count in total_counts:

                Z = landscape < total_count

                # get lengths of peptide sequences with coverage above the current threshold
                seqs_idx = np.where(np.diff(np.hstack(([False],~Z,[False]))))[0].reshape(-1,2)
                
                # get length of longest peptide subsequences with current count
                ce_start_pos = seqs_idx[np.diff(seqs_idx, axis=1).argmax(),0]
                current_pep_length = np.diff(seqs_idx, axis=1).max()
                
                # check if min_epi_length is fulfilled for that sequence
                if current_pep_length >= min_epi_len:

                    # get position of epitope in protein sequences
                    pep_in_prot_start = ce_start_pos
                    pep_in_prot_end = pep_in_prot_start + current_pep_length

                    # get consensus epitopes
                    whole_epitope_wo_mod = protein_df.at[r,'whole_epitopes'][group]
                    for _ in row['grouped_peptides_sequence'][group]:
                        protein_df.at[r,'consensus_epitopes_all'].append(whole_epitope_wo_mod[pep_in_prot_start:pep_in_prot_end])
                        protein_df.at[r,'core_epitopes_start_all'].append(pep_in_prot_start+min(row['grouped_peptides_start'][group]))
                        protein_df.at[r,'core_epitopes_end_all'].append(pep_in_prot_end+min(row['grouped_peptides_start'][group]) - 1) 
                    protein_df.at[r,'consensus_epitopes'].append(whole_epitope_wo_mod[pep_in_prot_start:pep_in_prot_end])
                    protein_df.at[r,'core_epitopes_start'].append(pep_in_prot_start+min(row['grouped_peptides_start'][group]))
                    protein_df.at[r,'core_epitopes_end'].append(pep_in_prot_end+min(row['grouped_peptides_start'][group]) - 1)
                    break
                
                # if no core with length > min_epi_length
                if total_count == total_counts[-1]:
                    pep_in_prot_start = ce_start_pos
                    pep_in_prot_end = pep_in_prot_start + current_pep_length
                    whole_epitope_wo_mod = protein_df.at[r,'whole_epitopes'][group]
                    for _ in row['grouped_peptides_sequence'][group]:
                        protein_df.at[r,'consensus_epitopes_all'].append(whole_epitope_wo_mod[pep_in_prot_start:pep_in_prot_end])
                        protein_df.at[r,'core_epitopes_start_all'].append(pep_in_prot_start+min(row['grouped_peptides_start'][group]))
                        protein_df.at[r,'core_epitopes_end_all'].append(pep_in_prot_end+min(row['grouped_peptides_start'][group]) - 1)
                    protein_df.at[r,'consensus_epitopes'].append(whole_epitope_wo_mod[pep_in_prot_start:pep_in_prot_end])
                    protein_df.at[r,'core_epitopes_start'].append(pep_in_prot_start+min(row['grouped_peptides_start'][group]))
                    protein_df.at[r,'core_epitopes_end'].append(pep_in_prot_end+min(row['grouped_peptides_start'][group]) - 1)
                    
    protein_df['proteome_occurrence'] = protein_df.apply(lambda row: [row['accession']+':'+str(row['consensus_epitopes_all'][i])+':'+str(row['core_epitopes_start_all'][i])+'-'+str(row['core_epitopes_end_all'][i]) for i in range(len(row['core_epitopes_start_all']))],axis=1)
    return protein_df


def reorder_peptides(row: pd.Series, intensity_column: str) -> pd.Series:
    """Reorder the peptides mapped to a protein by their position.
    
    Args:
        row: A row of a pandas dataframe containing per row one protein, 
            peptides mapped to the protein, start and end position of the peptide in the protein and the intensity of the peptide.
        intensity_column: The header of the column containing the intensities
            of the peptides.
    Returns:
        A reordered version of the input row, where the start positions are sorted in ascending order, the indices of the other columns are reordered in the same pattern. 
    """
    if intensity_column:
        lists = list(zip(row['start'], row['end'], row['sequence'], row['intensity']))
        sorted_lists = sorted(lists, key=lambda x: int(x[0]))
        starts, ends, sequences, intensities = zip(*sorted_lists)
        return list(starts), list(ends), list(sequences), list(intensities)
    else:
        lists = list(zip(row['start'], row['end'], row['sequence']))
        sorted_lists = sorted(lists, key=lambda x: int(x[0]))
        starts, ends, sequences = zip(*sorted_lists)
        return list(starts), list(ends), list(sequences)


def compute_consensus_epitopes(protein_df: pd.DataFrame, min_overlap: int, max_step_size: int, min_epi_len: int, intensity_column: float, mod_pattern: str) -> pd.DataFrame:
    """ Compute the core and whole sequence of all consensus epitope groups. 
    
    Args:
        protein_df: A pandas dataframe containing one protein per row.
        min_overlap: An integer of the minimal overlap between two epitopes     
            to be grouped to the same consensus epitope.
        max_step_size: An integer of the maximal distance between the start 
            position of two epitopes to be grouped to the same consensus 
            epitope.
        min_epi_len: An integer of the minimal length of a consensus epitope.
        intensity_column: The header of the column containing the intensities
            of the peptides.
        mod_pattern: A comma separated string with delimiters for peptide
            modifications

    Returns:
        The protein_df containing for each protein the core and whole sequence of each of its consensus epitope groups.
    """
    if intensity_column:
        protein_df[['start', 'end', 'sequence', 'intensity']] = protein_df.apply(lambda row: pd.Series(reorder_peptides(row, intensity_column)), axis=1)
    else:
        protein_df[['start', 'end', 'sequence']] = protein_df.apply(lambda row: pd.Series(reorder_peptides(row, intensity_column)), axis=1)
    # group peptides 
    protein_df = group_peptides(protein_df, min_overlap, max_step_size, intensity_column)
    protein_df = gen_landscape(protein_df,mod_pattern)
    protein_df = get_consensus_epitopes(protein_df, min_epi_len)
    return protein_df


"""
Reads in the evidence file and reference profile, computes the peptides positions in the proteome and links a protein accession with all peptides associated with the protein. 
"""

import pandas as pd 
import re
import numpy as np 
from Bio import SeqIO
import os 
import itertools

import logging
logger = logging.getLogger(__name__)


def read_id_output(id_output: str, seq_column: str, protacc_column: str, intensity_column: str, 
                   start_column: str, end_column: str, delimiter: str) -> pd.DataFrame:
    """Read in the evidence file.

    Args:
        id_output: The string of the path to the evidence file.
        seq_column: The string of the header of the column containing 
            peptide sequence information in the evidence file.
        protacc_column: The string of the header of the column containing 
            protein accession information in the evidence file.
        intensity_column: The string of the header of the column containing 
            intensity information in the evidence file.
        start_column: The string of the header of the column containing the 
            start positions of peptides in proteins.
        end_column: The string of the header of the column containing the end 
            position of peptides in proteins.
        delimiter: The delimiter that separates multiple entries in one column 
            in the evidence file.

    Returns:
        A pandas dataframe containing the columns sequence, protein accession, 
        peptides intensity and optional the columns start and end of the input 
        evidence file.

    Raises:
        Exception: If the file type of the provided evidence file is not 
            supported.
        Exception: If the provided column does not exist in the provided 
            evidence file.
        Exception: If a mandatory column header is not provided.
    """

    # determine the file type
    ext = os.path.splitext(id_output)[1]
    if ext == '.csv':
        peptides_df = pd.read_csv(id_output, delimiter=',')
    elif ext == '.tsv':
        peptides_df = pd.read_csv(id_output, delimiter='\t')
    elif ext == '.xlsx':
        peptides_df = pd.read_excel(id_output)
    else:
        raise Exception('The file type of your evidence file is not supported. \
                        Please use an evidence file that has one of the \
                        following file types: csv, tsv, xlsx')

    # check that the mandatory headers are provided and all provided column 
    # headers are part of the evidence file
    if seq_column not in peptides_df.columns:
        if seq_column:
            raise Exception('The header {} does not exist in the provided \
                            evidence file. Please provide the correct header.'.
                            format(seq_column))
        else:
            raise Exception('The header for the column containing the peptide \
                            sequences in the evidence file is mandatory. \
                            Please provide the correct header name.')
    if protacc_column not in peptides_df.columns:
        if protacc_column:
            raise Exception('The header {} does not exist in the provided \
                            evidence file. Please provide the correct header.'.
                            format(protacc_column))
        else:
            raise Exception('The header for the column containing the protein \
                            accessions in the evidence file is mandatory. \
                            Please provide the correct header name.')
    if (intensity_column not in peptides_df.columns) and intensity_column:
        raise Exception('The header {} does not exist in the provided \
                            evidence file. Please provide the correct header.'.
                            format(intensity_column))
    if (start_column not in peptides_df.columns) and start_column:
        raise Exception('The header {} does not exist in the provided \
                            evidence file. Please provide the correct header.'.
                            format(start_column))
    if (end_column not in peptides_df.columns) and end_column:
        raise Exception('The header {} does not exist in the provided \
                            evidence file. Please provide the correct header.'.
                            format(end_column))

    # read in evidence file (start, stop position and intensity only when 
    # provided)
    if start_column and end_column:
        if intensity_column:
            peptides_df = peptides_df[[protacc_column,seq_column,       
                                       intensity_column, start_column, 
                                       end_column]]  
        else:
            peptides_df = peptides_df[[protacc_column,seq_column, start_column, 
                                       end_column]]
        peptides_df = peptides_df.astype(str)
        # split if peptide occurs multiple times in proteome 
        peptides_df[start_column] = peptides_df[start_column].apply(
            lambda x: list(x.split(delimiter)))
        peptides_df[end_column] = peptides_df[end_column].apply(
            lambda x: list(x.split(delimiter)))
        peptides_df[protacc_column] = peptides_df[protacc_column].apply(
            lambda x: list(x.split(delimiter)))
    else:
        if intensity_column:
            peptides_df = peptides_df[[protacc_column,seq_column, intensity_column]]
        else:
            peptides_df = peptides_df[[protacc_column,seq_column]]
        # split accessions if peptide occurs multiple times in proteome
        peptides_df[protacc_column] = peptides_df[protacc_column].apply(lambda x: list(set(x.split(delimiter))))

    return peptides_df


def proteome_to_dict(proteome: str) -> dict[str,str]:
    """Read reference proteome into dictionary.

    Args:
        proteome: The string of the path to the reference proteome.

    Returns: 
        The reference proteome as a dictionary.
    """
    proteome_dict = {}
    proteome = SeqIO.parse(open(proteome),'fasta')
    for protein in proteome:
        proteome_dict[protein.id] = str(protein.seq)
    return proteome_dict



def compute_pep_pos(peptide: str, accession: str, proteome_dict: dict[str,str]) -> tuple[list[int],list[int]]:
    """Compute the position of a peptide in a proteome.

    Args:
        peptide: The string of a peptide sequence.
        accession: The string of a protein accession.
        proteome_dict: A dictionary containing the reference proteome.

    Returns:
        Returns a tuple (start,end), where start is a list of all start positions of the peptide in the protein and end is a list of all end positions of the peptide in the protein.
    """   
    
    prot_seq = proteome_dict[accession]

    # regular expression for all occurrences of peptide in protein (also overlapping ones) 
    peptide_search = '(?=' + peptide + ')' 
    
    # get all occurrences of the peptide in the protein 
    pos_list = [[match.start(),match.start()+len(peptide)-1] for match in re.finditer(peptide_search,prot_seq)]
    if len(pos_list) == 0:
        raise Exception('The peptide {} does not occur in the protein with accession "{}"! Please use the proteome that was used for the identification of the peptides.'.format(peptide, accession))
    
    pos_list = np.array(pos_list)
    start = pos_list[:,0].astype(str).tolist()
    end = pos_list[:,1].astype(str).tolist()

    return start, end


def group_repetitive(starts: list[int], ends: list[int], peptide: str, accession:str)->tuple[list[int],list[int]]:
    """Group peptide occurrences that belong to the same repetitive region.

    Args: 
        peptide: The string of a peptide sequence.
        accession: The string of a protein accession.
        starts: A list of all start positions of the peptide in the protein.
        ends: A list of all end positions of the peptide in the protein. 

    Returns:
        Returns a tuple (starts,ends), where starts is a list containing the
        updated start positions and ends is a list of updated end positions. 
        These contain from the input start and end positions all start and end 
        positions that are not part of a repetitive region. For the start and 
        end positions that are part of repetitive regions the lowest start 
        position and highest end position is kept for each repetitive region. 
    """
    updated_starts = []
    updated_ends = []
    # add the first occurrences start positions to the start positions
    updated_starts.append(starts[0])
    for pep_pos in range(len(starts)-1):
        # two start positions are not part of one repetitive region if the next start position is higher than the current end position 
        if starts[pep_pos + 1] > ends[pep_pos]:
            updated_starts.append(starts[pep_pos + 1])
            updated_ends.append(ends[pep_pos])
    # add the last occurrences end position to the end positions
    updated_ends.append(ends[-1])
    if len(updated_starts) < len(starts): 
        return updated_starts, updated_ends
    else:
        return starts, ends


def get_start_end(row: pd.Series, peptide: str) -> tuple[list[int],list[int],list[float]]:
    """Gets the starts, ends and intensity of a peptide.
    
    Args:
        row: A row of a pandas dataframe containing per row one protein, 
            peptides mapped to the protein with their start and end position 
            and their intensity. 
        peptide: The string of a peptide sequence.

    Returns: 
        Returns a triple (starts, ends, intensity), where starts are the start 
        positions of the peptide in the protein, ends are the end positions of 
        the peptide in the protein and intensity is the intensity of the 
        peptide in the protein.
    """
    #TODO add correct headers and add intensity
    all_starts = row['start']
    all_ends = row['end']
    #intensities = row['intensities']
    peptides = row['sequence']
    starts = []
    ends = []
    for n_pep in range(len(peptides)):
        if peptide == peptides[n_pep]:
            starts.append(all_starts[n_pep])
            ends.append(all_ends[n_pep])
    return starts, ends
            


def prot_pep_link(peptides_df: pd.DataFrame, seq_column: str, protacc_column: str, intensity_column: str, start_column: str, end_column: str, proteome_dict: dict[str,str], mod_pattern:str) -> pd.DataFrame:
    """Converts a dataframe from one peptide per row to one protein per row.
    
    Args:
        peptides_df: A pandas dataframe containing the columns peptide sequence,
            protein accession, intensity, start position and end position.
        seq_column: The string of the header of the column containing 
            peptide sequence information in the evidence file.
        protacc_column: The string of the header of the column containing 
            protein accession information in the evidence file.
        intensity_column: The string of the header of the column containing 
            intensity information in the evidence file.
        start_column: The string of the header of the column containing the 
            start positions of peptides in proteins.
        end_column: The string of the header of the column containing the end 
            position of peptides in proteins.
        proteome_dict: A dictionary containing the reference proteome.
        mod_pattern: A comma separated string with delimiters for peptide
            modifications
    
    Returns:
        A pandas dataframe containing one protein per row and all peptides 
        mapped to that protein in the peptides_df, with their start position, 
        end position and intensity.

    Raises:
        Exception: If a peptide does not occur in the protein to which it is 
            mapped.
    """
    if not start_column and not end_column:
        # if the start and end positions of the peptides is not defined in the input evidence file 
        

        if intensity_column:
            proteins = pd.DataFrame(columns=['accession', 'sequence', 'intensity'])
        else:
            proteins = pd.DataFrame(columns=['accession', 'sequence'])

        # create a row for each protein, that contains all peptides matched to the protein in the input evidence file
        for _, peptide in peptides_df.iterrows():
            # get all proteins matched to the peptide
            prot_accessions = peptide[protacc_column]
            for i, prot_accession in enumerate(prot_accessions):
                if prot_accession in proteins['accession'].values:
                    # update row for accessions seen before
                    proteins.loc[proteins['accession'] == prot_accession, 'sequence'] = proteins.loc[proteins['accession'] == prot_accession, 'sequence'].apply(lambda x: x + [peptide[seq_column]])
                    if intensity_column:
                        proteins.loc[proteins['accession'] == prot_accession, 'intensity'] = proteins.loc[proteins['accession'] == prot_accession, 'intensity'].apply(lambda x: x + [peptide[intensity_column]])

                else:
                    # create new row for accessions not seen before 
                    if intensity_column:
                        protein_entry = {'accession':prot_accession, 'sequence':[peptide[seq_column]], 'intensity':[peptide[intensity_column]]}
                    else:
                        protein_entry = {'accession':prot_accession, 'sequence':[peptide[seq_column]]}
                    proteins.loc[len(proteins)] = protein_entry

        # add the columns start and end
        proteins['start'] = [[] for _ in range(len(proteins))]
        proteins['end'] = [[] for _ in range(len(proteins))]

        # compute the peptide positions for each peptide of each protein
        for p, protein in proteins.iterrows():

            # get peptide sequence, protein accession and the measured intensity of the current row
            peptides = protein['sequence']
            accession = protein['accession']
            if intensity_column:
                intensities = protein['intensity']

            # starts and ends keep track of all occurrences of all peptides in peptides associated with the protein accession accession
            starts = []
            ends = []

            # updated_peps contains at each index the peptide associated with the start and end position in starts and ends at that index 
            # updated_intens contains at each index the intensity associated with the peptide at that index in updated_intens
            updated_peps = []
            if intensity_column:
                updated_intens = []

            for n_p, peptide in enumerate(peptides):
                # add each peptide and intensity associated with the accession to updated_peps and updated_intens
                updated_peps.append(peptide)
                if intensity_column:
                    updated_intens.append(intensities[n_p])
                
                # remove modifications in the sequence for the matching step (modifications are separated from the peptide by (), [] or by the user defined mod_pattern)
                pattern = r'\(.*?\)'
                peptide = re.sub(pattern,"",peptide)
                pattern = r'\[.*?\]'
                peptide = re.sub(pattern,"",peptide)
                if mod_pattern:
                    pattern = re.escape(mod_pattern.split(',')[0]) + r'.*?' + re.escape(mod_pattern.split(',')[1])
                    peptide = re.sub(pattern,"",peptide)
                pep_start, pep_end = compute_pep_pos(peptide, accession, proteome_dict)

                if len(pep_start) == 0:
                    raise Exception('The peptide {} does not occur in the protein with accession {} in the proteome you specified, but your input file provides evidence for that! Please use the proteome that was used for the identification of the peptides.'.format(peptide, prot_accession))
                if len(pep_start) > 1:
                    pep_start, pep_end = group_repetitive(pep_start,pep_end, peptide, accession)
                   
                if len(pep_start) > 1:
                    # if a peptide occurs multiple times in a protein
                    for _ in pep_start[:-1]:
                        updated_peps.append(peptide)
                        if intensity_column:
                            updated_intens.append(intensities[n_p])
                
                # collect all start and end positions of the peptide in the protein
                starts.extend(pep_start)
                ends.extend(pep_end)
            
            proteins.at[p, 'start'] = starts
            proteins.at[p, 'end'] = ends
            proteins.at[p, 'sequence'] = updated_peps
            if intensity_column:
                proteins.at[p,'intensity'] = updated_intens
    else:
        if intensity_column:
            proteins = pd.DataFrame(columns=['accession', 'sequence', 'intensity', 'start','end'])
        else:
            proteins = pd.DataFrame(columns=['accession', 'sequence', 'start','end'])

        for _, peptide in peptides_df.iterrows():
            # get all proteins associated with the peptide
            prot_accessions = peptide[protacc_column]
            for i, prot_accession in enumerate(prot_accessions):

                if prot_accession in proteins['accession'].values:
                    # append sequence, intensity, start and end information for accessions seen before
                    proteins.loc[proteins['accession'] == prot_accession, 'sequence'] = proteins.loc[proteins['accession'] == prot_accession, 'sequence'].apply(lambda x: x + [peptide[seq_column]])
                    if intensity_column:
                        proteins.loc[proteins['accession'] == prot_accession, 'intensity'] = proteins.loc[proteins['accession'] == prot_accession, 'intensity'].apply(lambda x: x + [peptide[intensity_column]])
                    proteins.loc[proteins['accession'] == prot_accession, 'start'] = proteins.loc[proteins['accession'] == prot_accession, 'start'].apply(lambda x: x + [peptide[start_column][i]])
                    proteins.loc[proteins['accession'] == prot_accession, 'end'] = proteins.loc[proteins['accession'] == prot_accession, 'end'].apply(lambda x: x + [peptide[end_column][i]])
                    
                else:
                    # create new row for accessions not seen before
                    if intensity_column:
                        protein_entry = {'accession':prot_accession, 'sequence':[peptide[seq_column]], 'intensity':[peptide[intensity_column]], 'start':[peptide[start_column][i]], 'end':[peptide[end_column][i]]}
                    else:
                        protein_entry = {'accession':prot_accession, 'sequence':[peptide[seq_column]], 'start':[peptide[start_column][i]], 'end':[peptide[end_column][i]]}
                    proteins.loc[len(proteins)] = protein_entry

        
        for p, protein in proteins.iterrows():

            # get peptide sequence, protein accession and the measured intensity of the current row
            peptides = protein['sequence']
            accession = protein['accession']
            if intensity_column:
                intensities = protein['intensity']
            starts = []
            ends = []
            
            # updated_peps contains at each index the peptide associated with the start and end position in starts and ends at that index 
            # updated_intens contains at each index the intensity associated with the peptide at that index in updated_intens
            updated_peps = []
            if intensity_column:
                updated_intens = []

            pep_pos_seen = []
            for n_p, peptide in enumerate(peptides):
                if peptide in pep_pos_seen:
                    continue
                else: 
                    pep_pos_seen.append(peptide)

                # add each peptide and intensity associated with the accession to updated_peps and updated_intens
                updated_peps.append(peptide)
                if intensity_column:
                    updated_intens.append(intensities[n_p])
                
                # get all starts, ends and the intensity of that peptide in that row
                pep_start, pep_end = get_start_end(protein, peptide)
            
                if len(pep_start) == 0:
                    raise Exception('The peptide {} does not occur in the protein with accession {} in the proteome you specified, but your input file provides evidence for that! Please use the proteome that was used for the identification of the peptides.'.format(peptide, prot_accession))
                if len(pep_start) > 1:
                    pep_start, pep_end = group_repetitive(pep_start,pep_end, peptide, accession)
                    
                   
                if len(pep_start) > 1:
                    # if a peptide occurs multiple times in a protein
                    for _ in pep_start[:-1]:
                        updated_peps.append(peptide)
                        if intensity_column:
                            updated_intens.append(intensities[n_p])
                
                # collect all start and end positions of the peptide in the protein
                starts.extend(pep_start)
                ends.extend(pep_end)
            

            proteins.at[p, 'start'] = starts
            proteins.at[p, 'end'] = ends
            proteins.at[p, 'sequence'] = updated_peps
            if intensity_column:
                proteins.at[p,'intensity'] = updated_intens
    return proteins


def parse_input(evidence_file: str, seq_column: str, protacc_column: str, intensity_column: str, start_column: str, end_column: str, delimiter: str, proteome_dict: dict[str,str], mod_pattern: str) -> pd.DataFrame:
    """Parse the evidence file.
    
    Args:
        evidence_file: The string of the path to the evidence file.
        seq_column: The string of the header of the column containing 
            peptide sequence information in the evidence file.
        protacc_column: The string of the header of the column containing 
            protein accession information in the evidence file.
        intensity_column: The string of the header of the column containing 
            intensity information in the evidence file.
        start_column: The string of the header of the column containing the 
            start positions of peptides in proteins.
        end_column: The string of the header of the column containing the end 
            position of peptides in proteins.
        delimiter: The delimiter that separates multiple entries in one column 
            in the evidence file.
        proteome_dict: A dictionary containing the reference proteome.
        mod_pattern: A comma separated string with delimiters for peptide
            modifications
    
    Returns:
        A tuple of a pandas dataframe and an integer. The dataframe contains 
        one protein per row and all peptides mapped to that protein in the 
        peptides_df, with their start position, end position and intensity.
        The integer is the number of peptides that are removed due to their 
        accession not appearing in the proteome file. 
    """
    peptides_df = read_id_output(evidence_file, seq_column, protacc_column, intensity_column, start_column, end_column, delimiter)

    # get peptides/proteins with protein accessions that do not appear in the proteome
    peptides = peptides_df.apply(lambda row: [prot for prot in row[protacc_column] if prot not in proteome_dict.keys()], axis=1).values
    n_removed_proteins = set(itertools.chain.from_iterable(peptides))

    # remove peptides with protein accessions that do not appear in the proteome 
    if start_column and end_column:
        peptides_df[start_column] = peptides_df.apply(lambda row: [start for start, prot in zip(row[start_column],row[protacc_column]) if prot in proteome_dict.keys()], axis=1)
        peptides_df[end_column] = peptides_df.apply(lambda row: [end for end, prot in zip(row[end_column],row[protacc_column]) if prot in proteome_dict.keys()], axis=1)
    peptides_df[protacc_column] = peptides_df.apply(lambda row: [prot for prot in row[protacc_column] if prot in proteome_dict.keys()], axis=1)

    logger.info(f'Peptides mapped to the following {len(n_removed_proteins)} proteins were removed since the proteins do not appear in the proteome fasta file: {n_removed_proteins}.')
    protein_df = prot_pep_link(peptides_df, seq_column, protacc_column, intensity_column, start_column, end_column, proteome_dict, mod_pattern)
    return protein_df, len(n_removed_proteins)



import os
import pandas as pd 
import ast
import yaml
import click
import logging
import numpy as np 
import matplotlib.pyplot as plt, mpld3
import warnings
import re

from . import __version__
from epicore_utils.modules.compute_cores import compute_consensus_epitopes
from epicore_utils.modules.map_result import map_pep_core, gen_epitope_df
from epicore_utils.modules.visualize_protein import plot_protein_landscape, plot_peptide_length_dist, plot_core_mapping_peptides_hist
from epicore_utils.modules.parse_input import parse_input, proteome_to_dict
from epicore_utils.modules.generate_report import gen_report

import logging

logger = logging.getLogger(__name__)

class InputParameter(object):
    """This class contains parameters necessary for the epicore script.

    Attributes:
        min_epi_length (int, optional): An integer of the minimal length of a 
            consensus epitope.
        min_overlap (int, optional): An integer of the minimal overlap between
            two epitopes to be grouped to the same consensus epitope.
        max_step_size (int, optional): An integer of the maximal distance 
            between the start position of two epitopes to be grouped to the same consensus epitope.
        seq_column (str, optional): The string of the header of the column 
            containing peptide sequence information in the evidence file.
        protacc_column (str, optional): The string of the header of the column  
            containing protein accession information in the evidence file.
        intensity_column (str, optional): The string of the header of the 
            column containing intensity information in the evidence file.
        delimiter (str, optional): The delimiter that separates multiple 
            entries in one column in the evidence file.
        mod_patter (str, optional): A comma separated string with delimiters
            for peptide modifications
        out_dir (str): A string of the directory, were all output files will be 
            saved.
        prot_accession (str, optional): A comma separated string containing the 
            protein accession, for which the protein landscape will be visualized.
        start_column (str, optional): The string of the header of the column 
            containing the start positions of peptides in proteins.
        end_column (str, optional): The string of the header of the column 
            containing the end position of peptides in proteins.

    """
    def __init__(self,reference_proteome=None, min_epi_length=None, min_overlap=None, max_step_size=None, seq_column=None, protacc_column=None, intensity_column=None, delimiter=None, mod_pattern=None, out_dir=None, prot_accession=None, start_column=None, end_column=None, report=None, html=None):
        self.min_epi_length = min_epi_length
        self.min_overlap = min_overlap
        self.max_step_size = max_step_size
        self.seq_column = seq_column
        self.protacc_column = protacc_column
        self.intensity_column = intensity_column
        self.delimiter = delimiter
        self.mod_pattern = mod_pattern
        self.out_dir = out_dir
        self.prot_accession = prot_accession
        self.start_column = start_column
        self.end_column = end_column
        self.report = report
        self.html = html
        self.proteome_dict = proteome_to_dict(reference_proteome)
        self.reference_proteome = reference_proteome

@click.version_option(__version__, "--version", "-V")

@click.group()
@click.option('--reference_proteome',type=click.Path(exists=True), required=True)
@click.option('--out_dir', type=click.Path(), required=True)
@click.pass_context
def main(ctx, reference_proteome, out_dir):
    ctx.obj = InputParameter(reference_proteome=reference_proteome, out_dir=out_dir)
    
@click.option('--min_epi_length', type=click.INT, default=11)
@click.option('--min_overlap', type=click.INT, default=11)
@click.option('--max_step_size', type=click.INT, required=5)
@click.option('--seq_column', type=click.STRING, required=True)
@click.option('--protacc_column', type=click.STRING, required=True)
@click.option('--intensity_column', type=click.STRING)
@click.option('--delimiter', type=click.STRING, required=True)
@click.option('--mod_pattern', type=click.STRING)
@click.option('--prot_accession', type=click.STRING)
@click.option('--start_column', type=click.STRING)
@click.option('--end_column', type=click.STRING)
@click.option('--report', is_flag=True)
@click.option('--html', is_flag=True)
@click.command()
@click.option('--evidence_file',type=click.Path(exists=True), required=True)
@click.pass_context
def generate_epicore_csv(ctx,evidence_file, min_epi_length, min_overlap, max_step_size, seq_column, protacc_column, intensity_column, delimiter, mod_pattern, prot_accession, start_column, end_column, report, html):
    ctx.obj = InputParameter(ctx.obj.reference_proteome, min_epi_length, min_overlap, max_step_size, seq_column, protacc_column, intensity_column, delimiter, mod_pattern, ctx.obj.out_dir, prot_accession, start_column, end_column, report, html)
    if not os.path.exists(ctx.obj.out_dir):
        os.mkdir(ctx.obj.out_dir)
    logging.basicConfig(filename=f'{ctx.obj.out_dir}/epicore.log', level=logging.INFO, format="%(asctime)s %(levelname)s %(message)s")
        
    # ----------------------
    #    Parse input file
    # ----------------------
    protein_df, n_removed_peps = parse_input(evidence_file, ctx.obj.seq_column, ctx.obj.protacc_column, ctx.obj.intensity_column, ctx.obj.start_column, ctx.obj.end_column, ctx.obj.delimiter, ctx.obj.proteome_dict, ctx.obj.mod_pattern)
    os.makedirs(ctx.obj.out_dir,exist_ok=True)

    # ----------------------
    # compute core epitopes, map peptides to cores
    # ----------------------
    protein_df = compute_consensus_epitopes(protein_df, ctx.obj.min_overlap, ctx.obj.max_step_size, ctx.obj.min_epi_length, ctx.obj.intensity_column, ctx.obj.mod_pattern)
    protein_df.to_csv(f'{ctx.obj.out_dir}/epicore_result.csv')
    pep_cores_mapping = map_pep_core(evidence_file,protein_df,ctx.obj.seq_column,ctx.obj.protacc_column,ctx.obj.start_column,ctx.obj.end_column,ctx.obj.intensity_column,ctx.obj.delimiter,ctx.obj.mod_pattern, ctx.obj.proteome_dict)
    pep_cores_mapping.to_csv(f'{ctx.obj.out_dir}/pep_cores_mapping.csv')


    # ----------------------
    # Reformat data and generate multiple plots
    # ----------------------
    # generate file with one epitope in each row
    epitope_df = gen_epitope_df(protein_df)
    epitope_df.to_csv(f'{ctx.obj.out_dir}/epitopes.csv')

    # compute length distribution of peptides and epitopes
    ext = os.path.splitext(evidence_file)[1]
    if ext == '.csv':
        evidence_df = pd.read_csv(evidence_file, delimiter=',')
    elif ext == '.tsv':
        evidence_df = pd.read_csv(evidence_file, delimiter='\t')
    elif ext == '.xlsx':
        evidence_df = pd.read_excel(evidence_file)
    evidence_df[ctx.obj.protacc_column] = evidence_df[ctx.obj.protacc_column].apply(lambda accessions: accessions.split(ctx.obj.delimiter))

    fig = plot_core_mapping_peptides_hist(epitope_df)
    if ctx.obj.html:
        fig.savefig(f'{ctx.obj.out_dir}/epitope_intensity_hist.svg')
        with open(f'{ctx.obj.out_dir}/epitope_intensity_hist.svg', 'r') as svg_file:
            svg_content = svg_file.read()
        svg_content = re.sub(r'<\?xml[^>]+\?>', '', svg_content)
        svg_content = re.sub(r'<!DOCTYPE[^>]+>', '', svg_content)
        html = f'<!DOCTYPE html> <html> <body>{svg_content}</body></html>'
        with open(f'{ctx.obj.out_dir}/epitope_intensity_hist.html','w') as f:
            f.write(html)
    else:
        fig.savefig(f'{ctx.obj.out_dir}/epitope_intensity_hist.svg')
    fig, peps, epitopes = plot_peptide_length_dist(evidence_df, epitope_df, ctx.obj.seq_column, 'whole_epitopes', ctx.obj.seq_column, 'whole_epitopes', 'peptides', 'whole epitopes')
    if ctx.obj.html:
        fig.savefig(f'{ctx.obj.out_dir}/length_distributions.svg')
        with open(f'{ctx.obj.out_dir}/length_distributions.svg', 'r') as svg_file:
            svg_content = svg_file.read()
        svg_content = re.sub(r'<\?xml[^>]+\?>', '', svg_content)
        svg_content = re.sub(r'<!DOCTYPE[^>]+>', '', svg_content)
        html = f'<!DOCTYPE html> <html> <body>{svg_content}</body></html>'
        with open(f'{ctx.obj.out_dir}/length_distributions.html','w') as f:
            f.write(html)
    else:
        fig.savefig(f'{ctx.obj.out_dir}/length_distributions.svg')
    
    # summarize some results
    if ctx.obj.report:
        gen_report(f'./{ctx.obj.out_dir}/length_distributions.svg', f'{ctx.obj.out_dir}/epitope_intensity_hist.svg', epitope_df, peps, epitopes, n_removed_peps, ctx, evidence_file,  f'{ctx.obj.out_dir}/epicore_result.csv')


@click.command()
@click.option('--epicore_csv',type=click.Path(exists=True), required=True)
@click.option('--protacc', type=click.STRING, required=True)
@click.pass_context
def plot_landscape(ctx,epicore_csv, protacc):
    if not protacc:
        raise Exception('No protein accession was provided. Please provide a protein accession')
    for accession in protacc.split(','):

        # read in precomputed protein coverage and epitope cores.
        protein_df = pd.read_csv(epicore_csv)

        protein_df['grouped_peptides_start'] = protein_df['grouped_peptides_start'].apply(ast.literal_eval)
        protein_df['core_epitopes_start'] = protein_df['core_epitopes_start'].apply(lambda cell: eval(cell, {"np": np}))
        protein_df['core_epitopes_end'] = protein_df['core_epitopes_end'].apply(lambda cell: eval(cell, {"np": np}))
        protein_df['landscape'] = protein_df['landscape'].apply(ast.literal_eval)

        if accession is not None:
            fig = plot_protein_landscape(protein_df,accession,ctx.obj.proteome_dict)
            fig.savefig(f'{ctx.obj.out_dir}/{accession}.pdf',bbox_inches='tight')
            

main.add_command(generate_epicore_csv)
main.add_command(plot_landscape)

if __name__ == '__main__':
    main()



"""
Assigns each peptide in the evidence files its core epitopes, the total intensity of that core epitope and the relative core intensity. 
"""
import warnings
import pandas as pd
warnings.simplefilter(action='ignore', category=pd.errors.PerformanceWarning)
import re
import os 
from itertools import repeat
import logging
import time
logger = logging.getLogger(__name__)

def read_entire_id_output(id_output: str) -> pd.DataFrame:
    """Read in the entire evidence file.
    
    Args:
        id_output: The string of the path to the evidence file.
    
    Returns:
        A pandas dataframe containing the evidence file.

    Raises:
        Exception: If the file type of the provided evidence file is not 
            supported.
    """
    # determine the file type
    ext = os.path.splitext(id_output)[1]
    if ext == '.csv':
        peptides_df = pd.read_csv(id_output, delimiter=',')
    elif ext == '.tsv':
        peptides_df = pd.read_csv(id_output, delimiter='\t')
    elif ext == '.xlsx':
        peptides_df = pd.read_excel(id_output)
    else:
        raise Exception('The file type of your evidence file is not supported. Please use an evidence file that has one of the following file types: csv, tsv, xlsx')
    return peptides_df

def map_pep_core(evidence_file: str, protein_df: pd.DataFrame, seq_column: str, protacc_column: str, start_column: str, end_column: str, intensity_column: str, delimiter: str, mod_pattern: str, proteome_dict: dict[str,str]) -> pd.DataFrame:
    """Map computed consensus epitope groups to the input evidence_file.
    
    Args:
        evidence_file: The string of the path to the evidence file.
        protein_df: A pandas dataframe containing one protein per row.
        seq_column: The string of the header of the column containing 
            peptide sequence information in the evidence file.
        protacc_column: The string of the header of the column containing 
            protein accession information in the evidence file.
        start_column: The string of the header of the column containing the 
            start positions of peptides in proteins.
        end_column: The string of the header of the column containing the end 
            position of peptides in proteins.
        intensity_column: The string of the header of the column containing 
            intensity information in the evidence file.
        delimiter: The delimiter that separates multiple entries in one column 
            in the evidence file.
        mod_pattern: A comma separated string with delimiters for peptide
            modifications

    Returns:
        The evidence_file with four additional columns containing the whole and 
        core sequence and total and relative intensity of each consensus 
        epitope group, to which the peptide of the row belongs.

    Raises:
        Exception: If the mappings are contradictory.
    """

    # read in entire evidence file
    evidence_file_df = read_entire_id_output(evidence_file)
    if intensity_column:
        protein_df = protein_df[['sequence', 'accession', 'start', 'end', 'intensity', 'whole_epitopes_all','consensus_epitopes_all', 'core_epitopes_intensity_all', 'relative_core_intensity_all', 'proteome_occurrence']]
    else:
        protein_df = protein_df[['sequence', 'accession', 'start', 'end', 'whole_epitopes_all','consensus_epitopes_all', 'proteome_occurrence']]

    # reformat protein_df so every peptide sequence is represented by one row
    if intensity_column:
        protein_df = protein_df.explode(['sequence', 'start', 'end', 'intensity', 'whole_epitopes_all','consensus_epitopes_all', 'core_epitopes_intensity_all', 'relative_core_intensity_all', 'proteome_occurrence'])
    else:
        protein_df = protein_df.explode(['sequence', 'start', 'end', 'whole_epitopes_all','consensus_epitopes_all', 'proteome_occurrence'])

    # reformat evidence file so each protein accession is represented by one row
    evidence_file_df[protacc_column] = evidence_file_df[protacc_column].str.split(delimiter)
    evidence_file_df = evidence_file_df.explode([protacc_column])
    evidence_merge_cols = [seq_column,protacc_column]
    if intensity_column:
        evidence_merge_cols.append(intensity_column)
    protein_merge_cols = ['sequence', 'accession']
    if 'intensity' in protein_df.columns:
        protein_merge_cols.append('intensity')

    # merge protein and evidence df to map each core epitope to the peptides that contribute to it 
    drop_cols = []
    evidence_file_df = evidence_file_df.merge(protein_df, left_on=evidence_merge_cols, right_on=protein_merge_cols)   
    if seq_column != 'sequence':
        drop_cols.append('sequence')
    if protacc_column != 'accession':
        drop_cols.append('accession')
    evidence_file_df = evidence_file_df[evidence_file_df.columns.drop(list(evidence_file_df.filter(regex='.*_y')))]
    evidence_file_df.columns = evidence_file_df.columns.str.replace(r'_x$','', regex=True)
    evidence_file_df = evidence_file_df.drop(drop_cols, axis=1)

    # group the rows which belong to the same peptide together (adds a list of all core epitopes belonging to that peptide)
    ev_cols = ['whole_epitopes_all', 'consensus_epitopes_all', protacc_column, 'proteome_occurrence']
    if intensity_column:
        ev_cols.append(intensity_column)
        ev_cols.append('core_epitopes_intensity_all')
        ev_cols.append('relative_core_intensity_all')
    evidence_file_df[ev_cols] = evidence_file_df[ev_cols].astype(str)
    # create dictionary for aggregation 
    all_cols = evidence_file_df.columns
    agg_dict = {}
    for col in all_cols:
        if col in [protacc_column,'whole_epitopes_all','consensus_epitopes_all', 'core_epitopes_intensity_all', 'relative_core_intensity_all', 'proteome_occurrence']:
            agg_dict[col] = lambda x: ','.join(x)
        else:
            agg_dict[col] = 'first'
    grouped_evidence_file_df = evidence_file_df.groupby([seq_column], as_index=False).agg(agg_dict)                                                                                                            
    return grouped_evidence_file_df

def gen_epitope_df(protein_df: pd.DataFrame) -> pd.DataFrame:
    """Generate dataframe that has one epitope per row.
    
    Args:
        protein_df: A pandas dataframe containing one protein per row.

    Returns:
        A reordered version of protein_df were each row stores one epitope.
    """
    # include intensity columns if present
    if ('core_epitopes_intensity' not in protein_df.columns) and ('relative_core_intensity' not in protein_df.columns):
        cols = ['whole_epitopes', 'consensus_epitopes','landscape', 'grouped_peptides_sequence']
    else:
        cols = ['whole_epitopes', 'consensus_epitopes','landscape', 'grouped_peptides_sequence', 'relative_core_intensity', 'core_epitopes_intensity']

    cols_acc = cols + ['accession']

    # separate each epitope in one row
    protein_df_long = protein_df.explode(cols)
    protein_df_long = protein_df_long.astype(str)
    epitopes_grouped_df = protein_df_long[cols_acc].groupby(cols)
    epitopes_grouped_df = epitopes_grouped_df.agg({'accession':lambda x:','.join(x)}).reset_index()
    
    logger.info(f'{len(epitopes_grouped_df)} unique epitopes were computed.')

    return epitopes_grouped_df
